import {App as BaseApp, AppOperations, Endpoint} from '../common';

interface BackendParams<App extends BaseApp> {
	app: App;
}

type AppOperationArgs<App extends BaseApp, Operation extends string> = (
	App extends BaseApp<string, { endpoints: any; operations: infer S }>
		? (
			S extends Record<string, readonly string[]>
				? S[Operation][number]
				: never
			)
		: never
	);

interface ImplementationFunctionParams<App extends BaseApp, Operation extends AppOperations<App> = AppOperations<App>> {
	endpoint: Endpoint;
	params: unknown;
	arg?: AppOperationArgs<App, Operation>;
	query?: URLSearchParams;
}

type ImplementationFunction<App extends BaseApp, Operation extends AppOperations<App> = AppOperations<App>> = (params: ImplementationFunctionParams<App, Operation>) => void;

export interface Backend<App extends BaseApp = BaseApp> {
	app: App;
	implementations: Map<string, ImplementationFunction<App>>;
	implementOperation<Operation extends AppOperations<App>>(operation: Operation, implementation: ImplementationFunction<App, Operation>): this;
}

class BackendInstance<App extends BaseApp> implements Backend<App> {
	readonly app: App;
	readonly implementations: Map<string, ImplementationFunction<App>>;

	constructor(params: BackendParams<App>) {
		this.app = params.app;
		this.implementations = new Map<string, ImplementationFunction<App>>();
	}

	implementOperation<Operation extends AppOperations<App>>(
		operation: Operation,
		implementation: ImplementationFunction<App, Operation>
	) {
		this.implementations.set(operation, implementation);
		return this;
	}
}

export const backend = <App extends BaseApp>(params: BackendParams<App>): Backend<App> => {
	return new BackendInstance(params);
};

export class PlainResponse {
	constructor() {
	}
}
