import {
	App as BaseApp,
	AVAILABLE_EXTENSION_METHODS,
	Endpoint, EndpointQueue,
	GetEndpointParams,
	Operation, serializeEndpointQueue,
	ServiceParams,
} from '../../common';
import {Client, ClientParams} from '../../client';

class ClientInstance<App extends BaseApp> implements Client<App> {
	readonly app: App;
	private readonly fetchFn: typeof fetch;
	private connection?: ServiceParams;
	private endpointQueue = [] as EndpointQueue;

	constructor(params: ClientParams<App>) {
		this.app = params.app;
		this.fetchFn = params.fetch ?? fetch;
	}

	connect(params: ServiceParams) {
		this.connection = params;
		return this;
	}

	at<TheEndpoint extends Endpoint = Endpoint>(endpoint: TheEndpoint, params?: Record<GetEndpointParams<TheEndpoint>, unknown>) {
		if (Array.isArray(this.endpointQueue)) {
			this.endpointQueue?.push([endpoint, params]);
		}
		return this;
	}

	makeRequest(operation: Operation) {
		const baseUrlFragments = [
			this.connection?.host ?? '0.0.0.0'
		];

		const thePort = (this.connection?.port ?? 80);
		if (thePort !== 80) {
			baseUrlFragments.push(thePort.toString());
		}

		const scheme = 'http';
		// todo need a way to decode url back to endpoint queue
		const url = serializeEndpointQueue(this.endpointQueue);
		this.endpointQueue = [];
		const rawEffectiveMethod = (operation.method ?? 'GET').toUpperCase();
		const finalEffectiveMethod = (AVAILABLE_EXTENSION_METHODS as unknown as string[]).includes(rawEffectiveMethod)
			? 'POST' as const
			: rawEffectiveMethod;

		return this.fetchFn(
			new URL(
				this.connection?.basePath ? `${this.connection.basePath}${url}` : url,
				`${scheme}://${baseUrlFragments.join(':')}`
			),
			{
				method: finalEffectiveMethod,
			},
		);
	}
}

export const client = <App extends BaseApp>(params: ClientParams<App>): Client<App> => {
	return new ClientInstance(params);
};
